#!/usr/bin/env python

import sys
import re
import optparse


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit()


def unescape(string):
    mapped_chars = {
        '>': '__gt__',
        '<': '__lt__',
        "'": '__sq__',
        '"': '__dq__',
        '[': '__ob__',
        ']': '__cb__',
        '{': '__oc__',
        '}': '__cc__',
        '@': '__at__',
        '\n': '__cn__',
        '\r': '__cr__',
        '\t': '__tc__',
        '#': '__pd__',
        '\\': 'X'   # Galaxy was replacing '\' with 'X' for some reason. Never found out why.
        }

    for key, value in mapped_chars.iteritems():
        string = string.replace(value, key)

    return string


def main():
    usage = """%prog [options]
options (listed below) default to 'None' if omitted
    """
    parser = optparse.OptionParser(usage=usage)

    parser.add_option(
        '-f', '--find',
        dest='find',
        metavar="STRING",
        default='',
        help='Find this string.')

    parser.add_option(
        '-r', '--replace',
        metavar="STRING",
        dest='replace',
        default='',
        help='Replace the found string with this.')

    parser.add_option(
        '-i', '--input',
        dest='input_txt',
        action='store',
        type='string',
        metavar="FILE",
        help='Name of file to be changed. STDIN is default')

    parser.add_option(
        '-e', '--regex',
        dest='regex',
        action='store_true',
        default=False,
        help='Treat the find input as regular expression.')

    parser.add_option(
        '-c', '--case',
        dest='case',
        action='store_true',
        default=False,
        help='Make the find case sensitive.')

    options, args = parser.parse_args()

    find = unescape(options.find)
    replace = unescape(options.replace)
    input_txt = unescape(options.input_txt)

    if not options.regex:
        find = re.escape(find)
    if options.case:
        insensitive_hippo = re.compile(find)
    else:
        insensitive_hippo = re.compile(find, re.IGNORECASE)

    with open(input_txt, 'r') as f:
            for line in f:
                newline = insensitive_hippo.sub(replace, line)
                print newline,

if __name__ == "__main__":
    main()
