#!/usr/bin/perl -w
use strict;
use Bio::TreeIO;
use Bio::Tree::Tree;

###this script will find the divergence time between two species
#input is a tree and 2 species or node names

# parse in newick/new hampshire format

my $infile = $ARGV[0];
my $multiplier = $ARGV[1];
my $outfile = $ARGV[2];

open(IN, "$infile") or exit;
open(OUT, ">$outfile") or exit;


my $tree = Bio::TreeIO->new(-format => 'newick',
					-file => $infile)->next_tree;


my $total_length = $tree->total_branch_length;
my @taxa = $tree->get_nodes;
my $ave_node_len = $total_length / @taxa;
my @leafs = $tree->get_leaf_nodes;

for my $node ( $tree->get_leaf_nodes ) {
	if($node->branch_length > ($multiplier * $ave_node_len)){
  		print OUT $node->id."\t".$ARGV[0]."\t".$node->branch_length."\t$ave_node_len\n";
  	}
}
my @LB_clade;
my @names;
my $newroot = $tree->get_root_node;
for my $node ( $tree->get_nodes ) {
	if($node->branch_length){
		if($node->branch_length > ($multiplier * $ave_node_len)){	
#This finds a long internal branch
			#Print descendants as Long Branch Clade
 			for my $child ( $node->get_all_Descendents ) {
				if($child->is_Leaf){
  					push(@LB_clade,$child->id."\t".$ARGV[0]."\t999\t$ave_node_len");
					push(@names, $child->id);
				}
			}
  		}
	}
}

if(@LB_clade > @leafs/2){	#More than half in LBA clade - remove those NOT in LBA clade
	for my $node ( $tree->get_leaf_nodes ) {
		my $curcheck = $node->id;
		chomp($curcheck);
		my $count = grep {/$curcheck/} @names;

		if($count==0){		#Tip was not collected in @names
	  		print OUT $node->id."\t".$ARGV[0]."\t999\t$ave_node_len\n";
		}
	}
}else{
	if(@LB_clade){
		print OUT join("\n",@LB_clade)."\n";
	}
}


close(IN);
